/* *******************************************************************
   Webserver

   How it all works together:

   page   0.htm         includes javascript j.js and f.css
   script j.js          the javascript requests the JSON
          json          returns data as JSON
   css    f.css         css for all on flash (program) memory
   php    c.php         not really php - only a resource doing actions and not returning content (just the http header 204 ok)
   ***************************************************************** */

void handleNotFound() {
  // Output a "404 not found" page. It includes the parameters which comes handy for test purposes.
  Serial.println(F("D015 handleNotFound()"));
  String message;
  message += F("404 - File Not Found\n"
               "URI: ");
  message += server.uri();
  message += F("\nMethod: ");
  message += (server.method() == HTTP_GET) ? "GET" : "POST";
  message += F("\nArguments: ");
  message += server.args();
  message += F("\n");
  for (uint8_t i = 0; i < server.args(); i++)
  {
    message += " " + server.argName(i) + ": " + server.arg(i) + "\n";
  }
  server.send(404, "text/plain", message);
}


void handle204()
{
  server.send(204);                // this page doesn't send back content
}


void addTop(String &message)
{
  message =  F("<!DOCTYPE html>\n"
               "<html lang='en'>\n"
               "<head>\n"
               "<title>" TXT_BOARDNAME " - Board " );
  message += USE_BOARD;
  message += F("</title>\n"
               "<meta http-equiv=\"content-type\" content=\"text/html; charset=utf-8\">\n"
               "<meta name=\"viewport\" content=\"width=device-width\">\n"
               "<link rel='stylesheet' type='text/css' href='/f.css'>\n"
               "<script src='j.js'></script>\n"
               "</head>\n");
  message += F("<body>\n");
  message += F("<header>\n<h1>" TXT_BOARDNAME " - Board " );
  message += USE_BOARD;
  message += F("</h1>\n"
               "<nav><p><a href=\"/\">[Home]</a> <a href=\"1.htm\">[Page&nbsp;1]</a> </p></nav>\n</header>\n"
               "<main>\n");
}


void addBottom(String &message) {
  message += F("</main>\n"
               "<footer>\n<p>");                 // The footer will display the uptime, the IP the version of the sketch and the compile date/time
  if (ss > 604800)
  {
    message += F("<span id='week'>");
    message +=  ((ss / 604800) % 52);
    message += F("</span> weeks ");
  }
  if (ss > 86400)
  {
    message += F("<span id='day'>");
    message += ((ss / 86400) % 7);
    message += F("</span> days ");
  }
  if (ss > 3600)
  {
    message += F("<span id='hour'>");
    message += ((ss / 3600) % 24);
    message += F("</span> hours ");
  }

  message += F("<span id='min'>");
  message += ((ss / 60) % 60);
  message += F("</span> minutes ");

  message += F("<span id='sec'>");
  message += (ss % 60);
  message += F("</span> seconds since startup | Version " VERSION " | IP: ");
  message += WiFi.localIP().toString();
  message += F(" | " __DATE__ " " __TIME__ "</p>\n</footer>\n</body>\n</html>");
  server.send(200, "text/html", message);
}


// the html output
// finally check your output if it is valid html: https://validator.w3.org
// *** HOME ***  0.htm
void handlePage()
{
  String message;
  addTop(message);

  /* Sliders for PWM pins */
  message += F("<article>\n"
               "<h2>PWM</h2>\n"                            // example how to switch/toggle an output
               "<p>Slide to change output.</p>\n");
  for (size_t i = 0; i < sizeof(pwmPin) / sizeof(pwmPin[0]); i++)
  {
    message += F("<p style='padding-left:10px'>");                         // MISSING - that's not nice ... to have a style in source...
    message += F("<input type='range' step='1' min='0' max='63' value='"); // the slider has only 0..63, but the function will send 0..1023
    message += pwmPin[i].value;
    message += F("' id='valuepwmPin"); // we need a unique ID to be able to assign values via JavaScript
    message += i;
    message += F("' onInput='firesmooth(this.id, this.value)'><span class='sliderlabel'>");
    message += pwmPin[i].label;
    message += F(" (");
    message += i;
    message += F(")</span></p>\n");
  }
  message += F("</article>\n");

  /* displaying nice ON/OFF rocket switches instead of checkboxes */
  message += F("<article>\n"
               "<h2>Switch</h2>\n"                         // example how to switch/toggle an output
               "<p>Toggle switches for on/off outputs</p>\n");
  for (size_t i = 0; i < sizeof(outputPin) / sizeof(outputPin[0]); i++)
  {
    message += F("<p><div class='toggle'><label><input type='checkbox' ");
    if (digitalRead(outputPin[i].pin) == outputPin[i].ON)
      message += F("value='checked' checked='checked'");
    message += F(" onclick='firecb(this.id)' id='cboutputPin");
    message += i;
    message += F("'><span class='slider'></span><span class='desc'>");
    message += outputPin[i].label;
    message += F(" (");
    message += i;
    message += F(")</span></label></div></p>\n");
  }
  message += F("</article>\n");

  message += F("<iframe name='i' style='display: none' ></iframe>\n");         // hack to keep the button press in the window

  addBottom(message);
  server.send(200, "text/html", message);
}

// *** Other ***  1.htm
// just some values and basic examples
void handlePage1()
{
  String message;
  addTop(message);

  /* output of current values */
  message += F("<article>\n"
               "<h2>Values (with update)</h2>\n");
  message += F("<p>Internal Voltage measured by ESP: <span id='internalVcc'>");  // example how to show values on the webserver
  message += ESP.getVcc();
  message += F("</span>mV</p>\n");
  for (size_t i = 0; i < sizeof(inputPin) / sizeof(inputPin[0]); i++)
  {
    message += F("<p>inputPin");
    message += i;
    message += F(" - ");
    message += inputPin[i].label;
    message += F(": <span id='inputPin");
    message += i;
    message += F("'>");
    message += digitalRead(inputPin[i].pin);
    message += F("</span> ");
    message += F("</p>\n");
  }
  for (size_t i = 0; i < sizeof(pwmPin) / sizeof(pwmPin[0]); i++)
  {
    message += F("<p>pwmPin");
    message += i;
    message += F(" - ");
    message += pwmPin[i].label;
    message += F(": <span id='pwmPin");
    message += i;
    message += F("'>");
    message += pwmPin[i].value;
    message += F("</span>");
    message += F("</p>\n");
  }
  for (size_t i = 0; i < sizeof(outputPin) / sizeof(outputPin[0]); i++)
  {
    message += F("<p>outputPin");
    message += i;
    message += F(" - ");
    message += outputPin[i].label;
    message += F(": <span id='outputPin");
    message += i;
    message += F("'>");
    if (digitalRead(outputPin[i].pin) == outputPin[i].ON)
      message += F("1");
    else
      message += F("0");
    message += F("</span>");
    message += F("</p>\n");
  }
  message += F("</article>\n");

  /* Simple HREF links styled as buttons to toggle the outputs */
  message += F("<article>\n"
               "<h2>Switch</h2>\n"                                             // example how to switch/toggle an output
               "<p>Example how to switch/toggle outputs, or to initiate actions. The buttons are 'fake' buttons and only styled by CSS. Click to toggle the output.</p>\n");
  for (size_t i = 0; i < sizeof(outputPin) / sizeof(outputPin[0]); i++)
  {
    message += F("<p><a class='btn' href='c.php?toggle=");
    message += i;
    message += F("' target='i'>&nbsp;");
    message += outputPin[i].label;
    message += F("&nbsp;</a></p>\n");
  }
  message += F("</article>\n");

  /* Simple HREF links styled as buttons to toggle the PWM outputs */
  message += F("<article>\n"
               "<h2>PWM Pins</h2>\n"                                           // example how to switch/toggle an output
               "<p>Click to toggle the PWM output (full on / full off)</p>\n");
  for (size_t i = 0; i < sizeof(pwmPin) / sizeof(pwmPin[0]); i++)
  {
    message += F("<p><a class='btn' href='c.php?pwmToggle=");
    message += i;
    message += F("' target='i'>&nbsp;");
    message += pwmPin[i].label;
    message += F("&nbsp;</a></p>\n");
  }
  message += F("</article>\n");

  message += F("<iframe name='i' style='display: none' ></iframe>\n");         // hack to keep the button press in the window

  addBottom(message);
  server.send(200, "text/html", message);
}

void handleCss()
{
  // output of stylesheet
  // this is a straight forward example how to generat a static page from program memory
  String message;
  message = F(":root{\n"
              "--no-col-main: #1279b1;\n"
              "--no-col-on: #28A745;\n"
              "--no-col-accent: orange;\n"
              "--no-col-warning: #ffc108;\n"
              "--no-col-danger: #DC3445;\n"
              "--no-col-bg: white;\n"
              "--no-col-bg0: #00487A;\n"
              "--no-col-bg1: #D0D0D0;\n"
              "--no-col-bg2: #F5F5F5;\n"
              "--no-col-bg5: silver;\n"
              "--no-col-bgm: #263238;\n"
              "--no-col-font: white;\n"
              "--no-col-font5: dimgrey;\n"
              "}\n"
              "* {font-family:sans-serif}\n"
              "body{margin:10px;background:var(--no-col-bg)}\n"
              "h1, h2{color:white;background:var(--no-col-main);text-align:center}\n"
              "h1{margin:1px;padding:5px;font-size:1.2em}\n"
              "h2{font-size:1.0em}\n"
              "h3{font-size:0.9em}\n"
              "a{text-decoration:none;color:dimgray;text-align:center}\n"
              "main{text-align:center}\n"
              "article{vertical-align:top;display:inline-block;margin:0.2em;padding:0.1em;border-style:solid;border-width:1px;border-color:var(--no-col-bg1);background-color:var(--no-col-bg2);width:20em;text-align:left}\n"
              "article h2{margin-top:0;padding-bottom:1px}\n"
              "section{margin-bottom:0.2em;clear:both}\n"
              "table{border-collapse:separate;border-spacing:0 0.2em}\n"
              "th, td{background-color:var(--no-col-bg1)}\n"
              "footer p{margin-bottom:5px;color:dimgray;background:var(--no-col-bg5);font-size:0.7em;text-align:center}\n"
              "nav{background-color:var(--no-col-bg5);margin:1px;padding:5px;font-size:0.8em}\n"
              "nav a{color:var(--no-col-font5);padding:10px;text-decoration:none}\n"
              "nav a:hover{text-decoration:underline}\n"
              "nav p{margin:0px;padding:0px}\n"
              "button, input[type='button'], input[type='submit'], .btn, .on, .off{margin-top:0;margin-bottom:0.2em;margin-left:4em;font-size:1.4em;background:var(--no-col-main);border-style:solid;border-width:1px;border-radius:.3rem;border-style:outset;width:5em;height:1.5em;color:var(--no-col-font);text-decoration:none;text-align:center}\n"
              ".on{background:var(--no-col-on)}\n"
              ".off{background:var(--no-col-danger)}\n"
              // checkbox toggle
              ".toggle label {position:relative;display:inline-block;width:8em;height:3.5em}\n"
              ".toggle input {display:none}\n"
              ".toggle .slider {position:absolute;cursor:pointer;top:1.5em;left:2em;width:4em;height:2em;background-color:darkgrey;border-radius:1em;transition:all .3s ease-in-out}\n"
              ".toggle .slider::before {position:absolute;content:\"\";height:1.6em;width:1.6em;left:0.2em;bottom:0.2em;background-color:white;border-radius:50%;transition:all .3s ease-in-out}\n"
              ".toggle input:checked + .slider {background-color:var(--no-col-on)}\n" // some kind of green/on
              ".toggle input:checked + .slider::before {transform:translateX(1.9em)}\n"
              ".cb input[type=checkbox]{visibility: hidden}\n"
              ".cb{width:25px;height:25px;background:#ddd;margin:10px 5px;position:relative;box-shadow:0px 1px 3px rgba(0,0,0,0.5);display:inline-block}\n"
              ".cb label{display:block;width:20px;height:18px;cursor:pointer;position:absolute;top:2px;left:2px;z-index:2;background:darkgrey;box-shadow:inset 0px 1px 3px rgba(0,0,0,0.5);color:white}\n"
              ".cb input[type=checkbox]:checked + label{background:#5a9900}\n"
              ".desc {width:200px;height:25px;margin:30px 120px;position:relative;display:inline-block}\n" // the description beside the switch
              // slider
              "input[type=range]{-webkit-appearance:none;margin: 12px 0px;/*width: 100%;*/}\n"
              "input[type=range]:focus{outline:none}\n"
              "input[type=range]::-webkit-slider-runnable-track{width:100%;height:8.4px;cursor:pointer;animate:0.2s;box-shadow: 1px 1px 1px #000000,0px 0px 1px #0d0d0d;background:dark;border-radius:1.3px;border: 0.2px solid #010101}\n"
              "input[type=range]::-webkit-slider-thumb {box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d;border:1px solid #000000;height:36px;width:16px;border-radius:3px;background:#F5F5F5;cursor:pointer;-webkit-appearance:none;margin-top: -14px}\n"
              "input[type=range]::-moz-range-track{width:100%;height:8.4px;cursor:pointer;animate:0.2s;box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d;background: dark;border-radius: 1.3px;border: 0.2px solid #010101}\n"
              "input[type=range]::-moz-range-thumb{box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d;border:1px solid #000000;height:36px;width:16px;border-radius:3px;background:#F5F5F5;cursor: pointer}\n"
              "input[type=range]::-ms-track{width: 100%;height: 8.4px;cursor: pointer;animate: 0.2s;background:transparent;border-color:transparent;border-width:16px 0;color:transparent}\n"
              "input[type=range]::-ms-fill-lower{background:darkgray;border: 0.2px solid #010101;border-radius: 2.6px;box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d}\n"
              "input[type=range]::-ms-fill-upper{background: dark;border:0.2px solid #010101;border-radius: 2.6px;box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d}\n"
              "input[type=range]::-ms-thumb {box-shadow: 1px 1px 1px #000000, 0px 0px 1px #0d0d0d;border: 1px solid #000000;height:36px;width:16px;background:#F5F5F5;cursor:pointer;border-radius: 3px}\n"
              ".sliderlabel{vertical-align:top;margin:13px 0 0 20px;position:relative;display:inline-block}\n"
             );
  server.send(200, "text/css", message);
}


void handleJson() {
  // Output: send data to browser as JSON
  // after modification always check if JSON is still valid. Just call the JSON (json) in your webbrowser and check.
  // Terminal.println(F("D295 requested json"));
  String message = "";
  message = (F(" {\"ss\":"));          // Start of JSON and the first object "ss":
  message += millis() / 1000;
  message += (F(",\"internalVcc\":"));
  message += ESP.getVcc();

  for (size_t i = 0; i < sizeof(inputPin) / sizeof(inputPin[0]); i++)
  {
    message += (F(",\"inputPin"));
    message += i;
    message += (F("\":"));
    message += digitalRead(inputPin[i].pin);
  }
  for (size_t i = 0; i < sizeof(outputPin) / sizeof(outputPin[0]); i++)
  {
    message += (F(",\"outputPin"));
    message += i;
    message += (F("\":"));
    if (digitalRead(outputPin[i].pin) == outputPin[i].ON)
      message += F("1");
    else
      message += F("0");
  }
  for (size_t i = 0; i < sizeof(pwmPin) / sizeof(pwmPin[0]); i++)
  {
    message += (F(",\"pwmPin"));
    message += i;
    message += (F("\":"));
    message += pwmPin[i].value;
  }
  message += (F("}"));                           // End of JSON
  server.send(200, "application/json", message); // set MIME type https://www.ietf.org/rfc/rfc4627.txt
}


void handleCommand() {
  // receive command and handle accordingly
  // Terminal.println(F("D324 handleCommand"));
  for (uint8_t i = 0; i < server.args(); i++) {
    Terminal.print(server.argName(i));
    Terminal.print(F(": "));
    Terminal.println(server.arg(i));
  }
  if (server.argName(0) == "toggle")             // the parameter which was sent to this server
  {
    byte i = server.arg(0).toInt();
    if (i < sizeof(outputPin) / sizeof(outputPin[0]))
    {
      if (digitalRead(outputPin[i].pin) != outputPin[i].ON)  // toggle: if the pin was on - switch it of and vice versa
      {                         
        digitalWrite(outputPin[i].pin, outputPin[i].ON);
      } else {
        digitalWrite(outputPin[i].pin, !outputPin[i].ON);
      }
    }
    else
    {
      Terminal.println(F("E352: invalid pin"));
    }
  }
  else if (server.argName(0) == "pwmToggle")     // the parameter which was sent to this server
  {
    // validieren?
    byte i = server.arg(0).toInt();
    if (i < sizeof(pwmPin) / sizeof(pwmPin[0]))
    {
      if (digitalRead(pwmPin[i].pin) == LOW) {   // toggle: if the pin was on - switch it of and vice versa
        digitalWrite(pwmPin[i].pin, HIGH);
        pwmPin[i].value = 1023;
      } else {
        digitalWrite(pwmPin[i].pin, LOW);
        pwmPin[i].value = 0;
      }
    }
    else
    {
      Terminal.println(F("E364: invalid pin"));
    }
  }
  else if (server.argName(0) == "range" && server.argName(1) == "v")  // a range based value from a slider
  {
    byte i = server.arg(0).substring(11).toInt();    // we get a "valuepwmPin"
    if (i < sizeof(pwmPin) / sizeof(pwmPin[0]))
    {
      //Terminal.print(F("D330 range i:")); Terminal.println(i);
      pwmPin[i].value = constrain(server.arg(1).toInt(), 0, 1023);
      //MISSING smooth
      analogWrite(pwmPin[i].pin, pwmPin[i].value);
    }
    else
    {
      Terminal.println(F("E379: invalid pin"));
    }
  }
  else if (server.argName(0) == "cb" && server.argName(1) == "v")  // a check base value from a slider
  {
    byte i = server.arg(0).substring(11).toInt();    // we get a "cboutputPin"
    //Terminal.print(F("D392 cb i:")); Terminal.println(i);
    if (i < sizeof(outputPin) / sizeof(outputPin[0]))
    {
      if (server.arg(1) != "0")                       // we get a string not an integer
        digitalWrite(outputPin[i].pin, outputPin[i].ON);
      else
        digitalWrite(outputPin[i].pin, !outputPin[i].ON);
    }
    else
    {
      Terminal.println(F("E402: invalid pin"));
    }
  }
  else if (server.argName(0) == "CMD" && server.arg(0) == "RESET")   // Example how to reset the module. Just send ?CMD=RESET
  {
    Terminal.println(F("D238 will reset"));
    ESP.restart();
  }

  server.send(204);                      // this page doesn't send back content --> 204
}


void handleJs() {
  // Output: a fetch API / JavaScript
  // a function in the JavaScript uses fetch API to request a JSON file from the webserver and updates the values on the page if the object names and ID are the same
  String message;
  message += F("const url ='json';\n"
               "function renew(){\n"
               " document.getElementById('sec').style.color = 'blue'\n"                                              // if the timer starts the request, the second gets blue
               " fetch(url)\n" // Call the fetch function passing the url of the API as a parameter
               " .then(response => {return response.json();})\n"
               " .then(jo => {\n"
               "   document.getElementById('sec').innerHTML = Math.floor(jo['ss'] % 60);\n"                         // example how to change a value in the HTML page
               "   for (var i in jo)\n"
               "   {\n"
               "    if (document.getElementById(i)) document.getElementById(i).innerHTML = jo[i];\n"                // as long as the JSON name fits to the HTML id, the value will be replaced
               "    if (document.getElementById('value'+i)) document.getElementById('value'+i).value = pwm2range(jo[i]);\n"    // in some cases we force to update the value
               //"    if (document.getElementByID('cb'+i)) if(jo[i] == 1) {document.getElementById('cb'+i).checked = true;} else {document.getElementById('cb'+i).checked = false;}\n"  // we force to update a check box - only idea - not not working
               "   }\n"
               // add other fields here (e.g. the delivered JSON name doesn't fit to the html id
               // finally, update the runtime
               "   if (jo['ss'] > 60) { document.getElementById('min').innerHTML = Math.floor(jo['ss'] / 60 % 60);}\n"
               "   if (jo['ss'] > 3600) {document.getElementById('hour').innerHTML = Math.floor(jo['ss'] / 3600 % 24);}\n"
               "   if (jo['ss'] > 86400) {document.getElementById('day').innerHTML = Math.floor(jo['ss'] / 86400 % 7);}\n"
               "   if (jo['ss'] > 604800) {document.getElementById('week').innerHTML = Math.floor(jo['ss'] / 604800 % 52);}\n"
               "   document.getElementById('sec').style.color = 'dimgray';\n"  // if everything was ok, the second will be grey again.
               " })\n"
               " .catch(function() {\n"                                        // this is where you run code if the server returns any errors
               "  document.getElementById('sec').style.color = 'red';\n"
               " });\n"
               "}\n"
               "document.addEventListener('DOMContentLoaded', renew, setInterval(renew, ");
  message += ajaxIntervall * 1000;
  message += F("));\n");

  // smooth fading 64 values : https://www.mikrocontroller.net/articles/LED-Fading
  message += F("const pwmtable = [0, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 4, 4, 5, 5, 6, 6, 7, 8, 9,\n"
               " 10, 11, 12, 13, 15, 17, 19, 21, 23, 26, 29, 32, 36, 40, 44, 49, 55,\n"
               " 61, 68, 76, 85, 94, 105, 117, 131, 146, 162, 181, 202, 225, 250,\n"
               " 279, 311, 346, 386, 430, 479, 534, 595, 663, 739, 824, 918, 1023];\n"
               // the smoothed fire function
               "function firesmooth(c,v){\n"      // c channel (mostly the id/self) | v myValue
               " v = pwmtable[v];\n"
               " var request = new XMLHttpRequest(); \n"
               " request.open('GET', 'c.php?range='+c+'&v='+v);\n"
               " request.send(null);\n"
               //" renew();\n"   // could be done, but fires very often
               "};\n"
               // a converter to get a from a pwmValue 0..1023 a range index 0..63
               "function pwm2range(v){\n"
               " for (i = 0; i < 64; i++){\n"
               "  if (v <= pwmtable[i]) return i;}\n"
               " return 1023;\n"
               "}\n");

  message += F("function firecb(c){\n"                                 // c caller (mostly the id/self)
               " var request=new XMLHttpRequest();\n"
               " var v=0;\n"                                           // if value is initialisied with 0 we get ...
               " if (document.getElementById(c).checked) v=1;\n"       // ... 1 or 0 if checkbox is checked (better than true or false)
               " request.open('GET','c.php?cb='+c+'&v='+v,true);\n"
               " request.send(null);\n"
               "}\n");

  server.send(200, "text/javascript", message);
}
